<?php

declare (strict_types=1);
/*
 * This file is part of PHP CS Fixer.
 *
 * (c) Fabien Potencier <fabien@symfony.com>
 *     Dariusz Rumiński <dariusz.ruminski@gmail.com>
 *
 * This source file is subject to the MIT license that is bundled
 * with this source code in the file LICENSE.
 */
namespace ps_metrics_module_v4_0_6\PhpCsFixer\Fixer\PhpTag;

use ps_metrics_module_v4_0_6\PhpCsFixer\AbstractFixer;
use ps_metrics_module_v4_0_6\PhpCsFixer\FixerDefinition\CodeSample;
use ps_metrics_module_v4_0_6\PhpCsFixer\FixerDefinition\FixerDefinition;
use ps_metrics_module_v4_0_6\PhpCsFixer\FixerDefinition\FixerDefinitionInterface;
use ps_metrics_module_v4_0_6\PhpCsFixer\Preg;
use ps_metrics_module_v4_0_6\PhpCsFixer\Tokenizer\Token;
use ps_metrics_module_v4_0_6\PhpCsFixer\Tokenizer\Tokens;
/**
 * Fixer for rules defined in PSR1 ¶2.1.
 *
 * @author Dariusz Rumiński <dariusz.ruminski@gmail.com>
 */
final class FullOpeningTagFixer extends AbstractFixer
{
    /**
     * {@inheritdoc}
     */
    public function getDefinition() : FixerDefinitionInterface
    {
        return new FixerDefinition('PHP code must use the long `<?php` tags or short-echo `<?=` tags and not other tag variations.', [new CodeSample('<?

echo "Hello!";
')]);
    }
    /**
     * {@inheritdoc}
     */
    public function getPriority() : int
    {
        // must run before all Token-based fixers
        return 98;
    }
    /**
     * {@inheritdoc}
     */
    public function isCandidate(Tokens $tokens) : bool
    {
        return \true;
    }
    /**
     * {@inheritdoc}
     */
    protected function applyFix(\SplFileInfo $file, Tokens $tokens) : void
    {
        $content = $tokens->generateCode();
        // replace all <? with <?php to replace all short open tags even without short_open_tag option enabled
        $newContent = Preg::replace('/<\\?(?:phP|pHp|pHP|Php|PhP|PHp|PHP)?(\\s|$)/', '<?php$1', $content, -1, $count);
        if (!$count) {
            return;
        }
        /* the following code is magic to revert previous replacements which should NOT be replaced, for example incorrectly replacing
         * > echo '<? ';
         * with
         * > echo '<?php ';
         */
        $newTokens = Tokens::fromCode($newContent);
        $tokensOldContentLength = 0;
        foreach ($newTokens as $index => $token) {
            if ($token->isGivenKind(\T_OPEN_TAG)) {
                $tokenContent = $token->getContent();
                if ('<?php' !== \strtolower(\substr($content, $tokensOldContentLength, 5))) {
                    $tokenContent = '<? ';
                }
                $tokensOldContentLength += \strlen($tokenContent);
                continue;
            }
            if ($token->isGivenKind([\T_COMMENT, \T_DOC_COMMENT, \T_CONSTANT_ENCAPSED_STRING, \T_ENCAPSED_AND_WHITESPACE, \T_STRING])) {
                $tokenContent = '';
                $tokenContentLength = 0;
                $parts = \explode('<?php', $token->getContent());
                $iLast = \count($parts) - 1;
                foreach ($parts as $i => $part) {
                    $tokenContent .= $part;
                    $tokenContentLength += \strlen($part);
                    if ($i !== $iLast) {
                        $originalTokenContent = \substr($content, $tokensOldContentLength + $tokenContentLength, 5);
                        if ('<?php' === \strtolower($originalTokenContent)) {
                            $tokenContent .= $originalTokenContent;
                            $tokenContentLength += 5;
                        } else {
                            $tokenContent .= '<?';
                            $tokenContentLength += 2;
                        }
                    }
                }
                $newTokens[$index] = new Token([$token->getId(), $tokenContent]);
                $token = $newTokens[$index];
            }
            $tokensOldContentLength += \strlen($token->getContent());
        }
        $tokens->overrideRange(0, $tokens->count() - 1, $newTokens);
    }
}
