<?php

/**
 * Checks that control structures have the correct spacing around brackets.
 *
 * @author    Greg Sherwood <gsherwood@squiz.net>
 * @copyright 2006-2015 Squiz Pty Ltd (ABN 77 084 670 600)
 * @license   https://github.com/PHPCSStandards/PHP_CodeSniffer/blob/master/licence.txt BSD Licence
 */
namespace ps_metrics_module_v4_0_6\PHP_CodeSniffer\Standards\PSR2\Sniffs\ControlStructures;

use ps_metrics_module_v4_0_6\PHP_CodeSniffer\Files\File;
use ps_metrics_module_v4_0_6\PHP_CodeSniffer\Sniffs\Sniff;
use ps_metrics_module_v4_0_6\PHP_CodeSniffer\Util\Tokens;
class ControlStructureSpacingSniff implements Sniff
{
    /**
     * How many spaces should follow the opening bracket.
     *
     * @var integer
     */
    public $requiredSpacesAfterOpen = 0;
    /**
     * How many spaces should precede the closing bracket.
     *
     * @var integer
     */
    public $requiredSpacesBeforeClose = 0;
    /**
     * Returns an array of tokens this test wants to listen for.
     *
     * @return array<int|string>
     */
    public function register()
    {
        return [\T_IF, \T_WHILE, \T_FOREACH, \T_FOR, \T_SWITCH, \T_ELSEIF, \T_CATCH, \T_MATCH];
    }
    //end register()
    /**
     * Processes this test, when one of its tokens is encountered.
     *
     * @param \PHP_CodeSniffer\Files\File $phpcsFile The file being scanned.
     * @param int                         $stackPtr  The position of the current token
     *                                               in the stack passed in $tokens.
     *
     * @return void
     */
    public function process(File $phpcsFile, $stackPtr)
    {
        $this->requiredSpacesAfterOpen = (int) $this->requiredSpacesAfterOpen;
        $this->requiredSpacesBeforeClose = (int) $this->requiredSpacesBeforeClose;
        $tokens = $phpcsFile->getTokens();
        if (isset($tokens[$stackPtr]['parenthesis_opener']) === \false || isset($tokens[$stackPtr]['parenthesis_closer']) === \false) {
            return;
        }
        $parenOpener = $tokens[$stackPtr]['parenthesis_opener'];
        $parenCloser = $tokens[$stackPtr]['parenthesis_closer'];
        $nextContent = $phpcsFile->findNext(\T_WHITESPACE, $parenOpener + 1, null, \true);
        if (\in_array($tokens[$nextContent]['code'], Tokens::$commentTokens, \true) === \false) {
            $spaceAfterOpen = 0;
            if ($tokens[$parenOpener + 1]['code'] === \T_WHITESPACE) {
                if (\strpos($tokens[$parenOpener + 1]['content'], $phpcsFile->eolChar) !== \false) {
                    $spaceAfterOpen = 'newline';
                } else {
                    $spaceAfterOpen = $tokens[$parenOpener + 1]['length'];
                }
            }
            $phpcsFile->recordMetric($stackPtr, 'Spaces after control structure open parenthesis', $spaceAfterOpen);
            if ($spaceAfterOpen !== $this->requiredSpacesAfterOpen) {
                $error = 'Expected %s spaces after opening bracket; %s found';
                $data = [$this->requiredSpacesAfterOpen, $spaceAfterOpen];
                $fix = $phpcsFile->addFixableError($error, $parenOpener + 1, 'SpacingAfterOpenBrace', $data);
                if ($fix === \true) {
                    $padding = \str_repeat(' ', $this->requiredSpacesAfterOpen);
                    if ($spaceAfterOpen === 0) {
                        $phpcsFile->fixer->addContent($parenOpener, $padding);
                    } else {
                        if ($spaceAfterOpen === 'newline') {
                            $phpcsFile->fixer->replaceToken($parenOpener + 1, '');
                        } else {
                            $phpcsFile->fixer->replaceToken($parenOpener + 1, $padding);
                        }
                    }
                }
            }
        }
        //end if
        $prev = $phpcsFile->findPrevious(\T_WHITESPACE, $parenCloser - 1, $parenOpener, \true);
        if ($tokens[$prev]['line'] === $tokens[$parenCloser]['line']) {
            $spaceBeforeClose = 0;
            if ($tokens[$parenCloser - 1]['code'] === \T_WHITESPACE) {
                $spaceBeforeClose = \strlen(\ltrim($tokens[$parenCloser - 1]['content'], $phpcsFile->eolChar));
            }
            $phpcsFile->recordMetric($stackPtr, 'Spaces before control structure close parenthesis', $spaceBeforeClose);
            if ($spaceBeforeClose !== $this->requiredSpacesBeforeClose) {
                $error = 'Expected %s spaces before closing bracket; %s found';
                $data = [$this->requiredSpacesBeforeClose, $spaceBeforeClose];
                $fix = $phpcsFile->addFixableError($error, $parenCloser - 1, 'SpaceBeforeCloseBrace', $data);
                if ($fix === \true) {
                    $padding = \str_repeat(' ', $this->requiredSpacesBeforeClose);
                    if ($spaceBeforeClose === 0) {
                        $phpcsFile->fixer->addContentBefore($parenCloser, $padding);
                    } else {
                        $phpcsFile->fixer->replaceToken($parenCloser - 1, $padding);
                    }
                }
            }
        }
        //end if
    }
    //end process()
}
//end class
